@@ -4,14 +4,17 @@
 # conda is distributed under the terms of the BSD 3-clause license.
 # Consult LICENSE.txt or http://opensource.org/licenses/BSD-3-Clause.
 '''
-These API functions have argument names referring to:
+We use the following conventions in this module:
 
-    dist:        canonical package name (e.g. 'numpy-1.6.2-py26_0')
+    dist:        canonical package name, e.g. 'numpy-1.6.2-py26_0'
 
-    PKGS_DIR:    the "packages directory" (e.g. '/opt/anaconda/pkgs')
+    ROOT_PREFIX: the prefix to the root environment, e.g. /opt/anaconda
 
-    PREFIX:      the prefix of a particular environment, which may also
-                 be the "default" environment (i.e. sys.prefix)
+    PKGS_DIR:    the "package cache directory", e.g. '/opt/anaconda/pkgs'
+                 this is always equal to ROOT_PREFIX/pkgs
+
+    prefix:      the prefix of a particular environment, which may also
+                 be the root environment
 
 Also, this module is directly invoked by the (self extracting) tarball
 installer to create the initial environment, therefore it needs to be
@@ -24,13 +27,13 @@ import sys
 import json
 import shutil
 import stat
-from os.path import dirname, exists, isdir, isfile, islink, join
+from os.path import abspath, dirname, exists, isdir, isfile, islink, join
 
 
 on_win = bool(sys.platform == 'win32')
 
 LINK_HARD = 1
-LINK_SOFT = 2
+LINK_SOFT = 2  # never used during the install process
 LINK_COPY = 3
 link_name_map = {
     LINK_HARD: 'hard-link',
@@ -39,8 +42,9 @@ link_name_map = {
 }
 
 # these may be changed in main()
-PREFIX = sys.prefix
-PKGS_DIR = join(PREFIX, 'pkgs')
+ROOT_PREFIX = sys.prefix
+PKGS_DIR = join(ROOT_PREFIX, 'pkgs')
+FORCE = False
 IDISTS = {}
 
 
@@ -50,8 +54,6 @@ def _link(src, dst, linktype=LINK_HARD):
 
     if linktype == LINK_HARD:
         os.link(src, dst)
-    elif linktype == LINK_SOFT:
-        os.symlink(src, dst)
     elif linktype == LINK_COPY:
         # copy relative symlinks as symlinks
         if islink(src) and not os.readlink(src).startswith('/'):
@@ -64,18 +66,18 @@ def _link(src, dst, linktype=LINK_HARD):
 
 def rm_rf(path):
     """
-    Completely delete path
+    try to delete path, but never fail
     """
-    if islink(path) or isfile(path):
-        # Note that we have to check if the destination is a link because
-        # exists('/path/to/dead-link') will return False, although
-        # islink('/path/to/dead-link') is True.
-        try:
+    try:
+        if islink(path) or isfile(path):
+            # Note that we have to check if the destination is a link because
+            # exists('/path/to/dead-link') will return False, although
+            # islink('/path/to/dead-link') is True.
             os.unlink(path)
-        except (OSError, IOError):
-            pass
-    elif isdir(path):
-        shutil.rmtree(path)
+        elif isdir(path):
+            shutil.rmtree(path)
+    except (OSError, IOError):
+        pass
 
 
 def yield_lines(path):
@@ -129,10 +131,10 @@ def exp_backoff_fn(fn, *args):
         except (OSError, IOError) as e:
             if e.errno in (errno.EPERM, errno.EACCES):
                 if n == max_tries - 1:
-                    raise
+                    raise Exception("max_tries=%d reached" % max_tries)
                 time.sleep(0.1 * (2 ** n))
             else:
-                raise
+                raise e
         else:
             return result
 
@@ -190,7 +192,7 @@ def name_dist(dist):
     return dist.rsplit('-', 2)[0]
 
 
-def create_meta(dist, info_dir, extra_info):
+def create_meta(prefix, dist, info_dir, extra_info):
     """
     Create the conda metadata, in a given prefix, for a given package.
     """
@@ -200,19 +202,19 @@ def create_meta(dist, info_dir, extra_info):
     # add extra info
     meta.update(extra_info)
     # write into <prefix>/conda-meta/<dist>.json
-    meta_dir = join(PREFIX, 'conda-meta')
+    meta_dir = join(prefix, 'conda-meta')
     if not isdir(meta_dir):
         os.makedirs(meta_dir)
     with open(join(meta_dir, dist + '.json'), 'w') as fo:
         json.dump(meta, fo, indent=2, sort_keys=True)
 
 
-def run_script(dist, action='post-link'):
+def run_script(prefix, dist, action='post-link'):
     """
     call the post-link (or pre-unlink) script, and return True on success,
     False on failure
     """
-    path = join(PREFIX, 'Scripts' if on_win else 'bin', '.%s-%s.%s' % (
+    path = join(prefix, 'Scripts' if on_win else 'bin', '.%s-%s.%s' % (
             name_dist(dist),
             action,
             'bat' if on_win else 'sh'))
@@ -229,7 +231,7 @@ def run_script(dist, action='post-link'):
         args = [shell_path, path]
 
     env = os.environ
-    env['ROOT_PREFIX'] = env['PREFIX'] = str(PREFIX)
+    env['PREFIX'] = prefix
 
     import subprocess
     try:
@@ -271,41 +273,29 @@ def read_no_link(info_dir):
     return res
 
 
-def try_hard_link(dist):
-    src = join(PKGS_DIR, dist, 'info', 'index.json')
-    dst = join(PREFIX, '.tmp-%s' % dist)
-    assert isfile(src), src
-    assert not isfile(dst), dst
-    try:
-        _link(src, dst, LINK_HARD)
-        return True
-    except OSError:
-        return False
-    finally:
-        rm_rf(dst)
-
-
-def linked():
+def linked(prefix):
     """
     Return the (set of canonical names) of linked packages in prefix.
     """
-    meta_dir = join(PREFIX, 'conda-meta')
+    meta_dir = join(prefix, 'conda-meta')
     if not isdir(meta_dir):
         return set()
     return set(fn[:-5] for fn in os.listdir(meta_dir) if fn.endswith('.json'))
 
 
-def link(dist, linktype=LINK_HARD):
+def link(prefix, dist, linktype=LINK_HARD):
     '''
-    Set up a package in a specified (environment) prefix.  We assume that
-    the package has been extracted (using extract() above).
+    Link a package in a specified prefix.  We assume that the packacge has
+    been extra_info in either
+      - <PKGS_DIR>/dist
+      - <ROOT_PREFIX>/ (when the linktype is None)
     '''
     if linktype:
         source_dir = join(PKGS_DIR, dist)
         info_dir = join(source_dir, 'info')
         no_link = read_no_link(info_dir)
     else:
-        info_dir = join(PREFIX, 'info')
+        info_dir = join(prefix, 'info')
 
     files = list(yield_lines(join(info_dir, 'files')))
     has_prefix_files = read_has_prefix(join(info_dir, 'has_prefix'))
@@ -313,12 +303,15 @@ def link(dist, linktype=LINK_HARD):
     if linktype:
         for f in files:
             src = join(source_dir, f)
-            dst = join(PREFIX, f)
+            dst = join(prefix, f)
             dst_dir = dirname(dst)
             if not isdir(dst_dir):
                 os.makedirs(dst_dir)
             if exists(dst):
-                rm_rf(dst)
+                if FORCE:
+                    rm_rf(dst)
+                else:
+                    raise Exception("dst exists: %r" % dst)
             lt = linktype
             if f in has_prefix_files or f in no_link or islink(src):
                 lt = LINK_COPY
@@ -330,12 +323,12 @@ def link(dist, linktype=LINK_HARD):
     for f in sorted(has_prefix_files):
         placeholder, mode = has_prefix_files[f]
         try:
-            update_prefix(join(PREFIX, f), PREFIX, placeholder, mode)
+            update_prefix(join(prefix, f), prefix, placeholder, mode)
         except PaddingError:
             sys.exit("ERROR: placeholder '%s' too short in: %s\n" %
                      (placeholder, dist))
 
-    if not run_script(dist, 'post-link'):
+    if not run_script(prefix, dist, 'post-link'):
         sys.exit("Error: post-link failed for: %s" % dist)
 
     meta = {
@@ -349,7 +342,7 @@ def link(dist, linktype=LINK_HARD):
     except KeyError:
         meta.update(read_urls(dist))
     #meta['installed_by'] = ...
-    create_meta(dist, info_dir, meta)
+    create_meta(prefix, dist, info_dir, meta)
 
 
 def duplicates_to_remove(linked_dists, keep_dists):
@@ -383,81 +376,75 @@ def duplicates_to_remove(linked_dists, keep_dists):
     return sorted(res)
 
 
-def link_idists(verbose=False):
+def link_idists():
+    raise NotImplementedError
+
+
+def prefix_env(env_name):
+    if env_name == 'root':
+        return ROOT_PREFIX
+    else:
+        return join(ROOT_PREFIX, 'envs', env_name)
+
+
+def post_extract(env_name='root'):
     """
-    link all distributions listed in the global IDISTS
+    assuming that the package is extracted in the environment `env_name`,
+    this function does everything link() does except the actual linking,
+    i.e. update prefix files, run 'post-link', creates the conda metadata,
+    and removed the info/ directory afterwards.
     """
-    if not IDISTS:
-        sys.exit("Error: invalid mode, maybe --post is missing?")
-
-    if on_win:
-        raise NotImplementedError
-
-    idists = sorted(IDISTS)
-    linktype = (LINK_HARD if try_hard_link(idists[0]) else LINK_COPY)
-    if verbose:
-        print("linktype: %s" % link_name_map[linktype])
-
-    for dist in idists:
-        if verbose:
-            print("linking: %s" % dist)
-        link(dist, linktype)
-
-    for dist in duplicates_to_remove(linked(), idists):
-        meta_path = join(PREFIX, 'conda-meta', dist + '.json')
-        print("WARNING: unlinking: %s" % meta_path)
-        try:
-            os.rename(meta_path, meta_path + '.bak')
-        except OSError:
-            rm_rf(meta_path)
-
-
-def post_extract():
-    """
-    assuming that the package is extracted in prefix itself, this function
-    does everything link() does except the actual linking, i.e.
-    update prefix files, run 'post-link', creates the conda metadata
-    """
-    with open(join(PREFIX, 'info', 'index.json')) as fi:
+    prefix = prefix_env(env_name)
+    info_dir = join(prefix, 'info')
+    with open(join(info_dir, 'index.json')) as fi:
         meta = json.load(fi)
     dist = '%(name)s-%(version)s-%(build)s' % meta
-    link(dist, linktype=None)
+    link(prefix, dist, linktype=None)
+    shutil.rmtree(info_dir)
 
 
 def main():
-    global PREFIX, PKGS_DIR
+    global ROOT_PREFIX, PKGS_DIR, FORCE
 
     from optparse import OptionParser
 
-    p = OptionParser(description="conda link tool used by installer")
+    p = OptionParser(description="conda link tool used by installers")
 
-    p.add_option('--prefix',
+    p.add_option('--root-prefix',
                  action="store",
-                 default=sys.prefix,
-                 help="prefix (defaults to %default)")
+                 default=abspath(join(__file__, '..', '..')),
+                 help="root prefix (defaults to %default)")
 
     p.add_option('--post',
-                 action="store_true",
-                 help="perform post extract (on a single package)")
-
-    p.add_option('-v', '--verbose',
-                 action="store_true")
+                 action="store",
+                 help="perform post extract (on a single package), "
+                      "in environment NAME",
+                 metavar='NAME')
 
     opts, args = p.parse_args()
     if args:
         p.error('no arguments expected')
 
-    PREFIX = opts.prefix
-    PKGS_DIR = join(PREFIX, 'pkgs')
-    if opts.verbose:
-        print("PREFIX: %r" % PREFIX)
+    ROOT_PREFIX = opts.root_prefix
+    PKGS_DIR = join(ROOT_PREFIX, 'pkgs')
 
     if opts.post:
-        post_extract()
+        post_extract(opts.post)
         return
 
-    link_idists(opts.verbose)
+    try:
+        FORCE = bool(int(os.getenv('FORCE', 0)))
+    except ValueError:
+        FORCE = False
+
+    if FORCE:
+        print("using -f (force) option")
+
+    link_idists()
 
 
 if __name__ == '__main__':
-    main()
+    if IDISTS:
+        main()
+    else: # common usecase
+        post_extract()
